/*
 * msmtp.c
 *
 * This file is part of msmtp, an SMTP client.
 *
 * Copyright (C) 2000, 2003, 2004, 2005
 * Martin Lambers <marlam@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *   msmtp is released under the GPL with the additional exemption that
 *   compiling, linking, and/or using OpenSSL is allowed.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
extern int errno;
#include <time.h>
#include <getopt.h>
extern char *optarg;
extern int optind;
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_SYSEXITS_H
#include <sysexits.h>
#else
#define EX_OK		0
#define EX_USAGE	64
#define EX_DATAERR	65
#define EX_NOINPUT	66
#define EX_NOUSER	67
#define EX_NOHOST	68
#define EX_UNAVAILABLE	69
#define EX_SOFTWARE	70
#define EX_OSERR	71
#define EX_OSFILE	72
#define EX_CANTCREAT	73
#define EX_IOERR	74
#define EX_TEMPFAIL	75
#define EX_PROTOCOL	76
#define EX_NOPERM	77
#define EX_CONFIG	78
#endif

#include "list.h"
#include "paths.h"
#include "conf.h"
#include "net.h"
#include "smtp.h"
#ifdef HAVE_SSL
#include "tls.h"
#endif /* HAVE_SSL */
#include "xmalloc.h"

/* Default configuration file. */
#ifdef WINDOWS
#define CONFFILE "msmtprc.txt"
#elif defined (DJGPP)
#define CONFFILE "_msmtprc"
#else /* UNIX */
#define CONFFILE ".msmtprc"
#endif

/* The size of all error message buffers */
#define ERRSTR_BUFSIZE 512
const size_t errstr_bufsize = ERRSTR_BUFSIZE;

/* The name of this program */
char *prgname;


/*
 * Translate error codes from net.h, tls.h or smtp.h 
 * to error codes from sysexits.h
 */

int exitcode_net(int net_error_code)
{
    switch (net_error_code)
    {
	case NET_EHOSTNOTFOUND:
	    return EX_NOHOST;

	case NET_ESOCKET:
	    return EX_OSERR;

	case NET_ECONNECT:
	    return EX_TEMPFAIL;	    

	case NET_EIO:
	    return EX_IOERR;
	    
	case NET_ELIBFAILED:
	default:
	    return EX_SOFTWARE;
    }
}

#ifdef HAVE_SSL
int exitcode_tls(int tls_error_code)
{
    switch (tls_error_code)
    {
	case TLS_EIO:
	    return EX_IOERR;

	case TLS_EFILE:
	    return EX_NOINPUT;

	case TLS_EHANDSHAKE:
	    return EX_PROTOCOL;

	case TLS_ECERT:
	    /* did not find anything better... */
	    return EX_UNAVAILABLE;

	case TLS_ELIBFAILED:
	case TLS_ESEED:
	default:
	    return EX_SOFTWARE;
    }
}
#endif /* HAVE_SSL */

int exitcode_smtp(int smtp_error_code)
{
    switch (smtp_error_code)
    {
	case SMTP_EIO:
	    return EX_IOERR;

	case SMTP_EPROTO:
	    return EX_PROTOCOL;

	case SMTP_EINVAL:
	    return EX_DATAERR;
	    
	case SMTP_EAUTHFAIL:
	    return EX_NOPERM;

	case SMTP_EINSECURE:
	case SMTP_EUNAVAIL:
	    return EX_UNAVAILABLE;

	case SMTP_ELIBFAILED:
	default:
	    return EX_SOFTWARE;
    }
}


/* 
 * Return the name of a sysexits.h exitcode
 */
const char *exitcode_to_string(int exitcode)
{
    switch (exitcode)
    {
	case EX_OK:
	    return "EX_OK";

	case EX_USAGE:
	    return "EX_USAGE";
	    
	case EX_DATAERR:
	    return "EX_DATAERR";

	case EX_NOINPUT:
	    return "EX_NOINPUT";

	case EX_NOUSER:
	    return "EX_NOUSER";

	case EX_NOHOST:
	    return "EX_NOHOST";

	case EX_UNAVAILABLE:
	    return "EX_UNAVAILABLE";

	case EX_SOFTWARE:
	    return "EX_SOFTWARE";

	case EX_OSERR:
	    return "EX_OSERR";

	case EX_OSFILE:
	    return "EX_OSFILE";
	    
	case EX_CANTCREAT:
	    return "EX_CANTCREAT";
	    
	case EX_IOERR:
	    return "EX_IOERR";

	case EX_TEMPFAIL:
	    return "EX_TEMPFAIL";
	    
	case EX_PROTOCOL:
	    return "EX_PROTOCOL";

	case EX_NOPERM:
	    return "EX_NOPERM";

	case EX_CONFIG:
	    return "EX_CONFIG";

	default:
	    return "BUG:UNKNOWN";
    }
}


/* 
 * msmtp_sanitize_string()
 *
 * Replaces all non-printable characters and all single quotes in the string
 * with an underscore.
 */

void msmtp_sanitize_string(char *s)
{
    while (*s != '\0')
    {
	if (!isprint((unsigned char)*s) || *s == '\'')
	{
    	    *s = '_';
	}
	s++;
    }
}


/*
 * msmtp_print_tls_cert_info()
 *
 * Prints information about a TLS certificate.
 */

#ifdef HAVE_SSL
void msmtp_print_tls_cert_info(tls_cert_info_t *tci)
{
    const char *info_fieldname[6] = { "Common Name", "Organization", 
	"Organizational unit", "Locality", "State or Province", "Country" };
    char hex[] = "0123456789ABCDEF";
    char sha1_fingerprint_string[60];
    char md5_fingerprint_string[48];
    char *tmp;
    int i;
    
    for (i = 0; i < 20; i++)
    {
	sha1_fingerprint_string[3 * i] = hex[(tci->sha1_fingerprint[i] & 0xf0) >> 4];
	sha1_fingerprint_string[3 * i + 1] = hex[tci->sha1_fingerprint[i] & 0x0f];
	sha1_fingerprint_string[3 * i + 2] = ':';
    }
    sha1_fingerprint_string[59] = '\0';
    for (i = 0; i < 16; i++)
    {
	md5_fingerprint_string[3 * i] = hex[(tci->md5_fingerprint[i] & 0xf0) >> 4];
	md5_fingerprint_string[3 * i + 1] = hex[tci->md5_fingerprint[i] & 0x0f];
	md5_fingerprint_string[3 * i + 2] = ':';
    }
    md5_fingerprint_string[47] = '\0';
    
    printf("TLS certificate information:\n");
    printf("  Owner:\n");
    for (i = 0; i < 6; i++)
    {
	if (tci->owner_info[i])
	{
	    tmp = xstrdup(tci->owner_info[i]);
	    msmtp_sanitize_string(tmp);
	    printf("    %s: %s\n", info_fieldname[i], tmp);
	    free(tmp);
	}
    }
    printf("  Issuer:\n");
    for (i = 0; i < 6; i++)
    {
	if (tci->issuer_info[i])
	{
	    tmp = xstrdup(tci->issuer_info[i]);
	    msmtp_sanitize_string(tmp);
	    printf("    %s: %s\n", info_fieldname[i], tmp);
	    free(tmp);
	}
    }
    printf("  Validity:\n");
    printf("    Activation time: %s", ctime(&tci->activation_time));
    printf("    Expiration time: %s", ctime(&tci->expiration_time));
    printf("  Fingerprints:\n");
    printf("    SHA1: %s\n", sha1_fingerprint_string);
    printf("    MD5:  %s\n", md5_fingerprint_string);
}
#endif


/*
 * msmtp_endsession()
 *
 * Quit an SMTP session and close the connection.
 * QUIT is only sent when the flag 'quit' is set.
 */

void msmtp_endsession(smtp_server_t *srv, int quit)
{
    char errstr[ERRSTR_BUFSIZE];
    
    if (quit)
    {
	(void)smtp_quit(srv, errstr);
    }
    smtp_close(srv);
}


/*
 * msmtp_rmqs()
 *
 * Sends an ETRN request to the SMTP server specified in the account 'acc'.
 * 'errstr' must be 'errstr_bufsize' characters long.
 * If an error occured, 'errstr' may contain a descriptive message (or an empty
 * string) and 'msg' may contain the offending message from the SMTP server (or
 * NULL).
 */

int msmtp_rmqs(account_t *acc, int debug, const char *rmqs_argument, 
	list_t **msg, char *errstr)
{
    smtp_server_t srv;
    int e;
#ifdef HAVE_SSL
    tls_cert_info_t *tci = NULL;
#endif /* HAVE_SSL */
    
    errstr[0] = '\0';
    *msg = NULL;
    
    /* create a new smtp_server_t */
    srv = smtp_new(debug ? stdout : NULL);

    /* prepare tls */
#ifdef HAVE_SSL
    if (acc->tls)
    {
	if ((e = smtp_tls_init(&srv, acc->tls_key_file, acc->tls_cert_file, 
			acc->tls_trust_file, errstr)) != TLS_EOK)
	{
	    return exitcode_tls(e);
	}
    }
#endif /* HAVE_SSL */

    /* connect */
    if ((e = smtp_connect(&srv, acc->host, acc->port, errstr)) != NET_EOK)
    {
	return exitcode_net(e);
    }

    /* start tls for ssmtp servers */
#ifdef HAVE_SSL
    if (acc->tls && acc->tls_nostarttls)
    {
	if (debug)
	{
	    tci = tls_cert_info_new();
	}
	if ((e = smtp_tls(&srv, acc->host, acc->tls_nocertcheck, tci, errstr)) 
		!= TLS_EOK)
	{
	    if (debug)
	    {
		tls_cert_info_free(tci);
	    }
	    msmtp_endsession(&srv, 0);
	    return exitcode_tls(e);
	}
	if (debug)
	{
	    msmtp_print_tls_cert_info(tci);
	    tls_cert_info_free(tci);
	}
    }
#endif /* HAVE_SSL */

    /* get greeting */
    if ((e = smtp_get_greeting(&srv, msg, NULL, errstr)) != SMTP_EOK)
    {
	msmtp_endsession(&srv, 0);
	return exitcode_smtp(e);
    }
		    
    /* initialize session */
    if ((e = smtp_init(&srv, acc->domain, msg, errstr)) != SMTP_EOK)
    {
	msmtp_endsession(&srv, 1);
	return exitcode_smtp(e);
    }
    
    /* start tls for starttls servers */
#ifdef HAVE_SSL
    if (acc->tls && !acc->tls_nostarttls)
    {
	if (!(srv.cap.flags & SMTP_CAP_STARTTLS))
	{
	    snprintf(errstr, errstr_bufsize,
		    "the SMTP server does not support TLS via the STARTTLS command");
	    msmtp_endsession(&srv, 1);
	    return EX_UNAVAILABLE;
	}
	if ((e = smtp_tls_starttls(&srv, msg, errstr)) != SMTP_EOK)
	{
	    msmtp_endsession(&srv, 1);
	    return exitcode_smtp(e);
	}
	if (debug)
	{
	    tci = tls_cert_info_new();
	}
	if ((e = smtp_tls(&srv, acc->host, acc->tls_nocertcheck, tci, errstr)) 
		!= TLS_EOK)
	{
	    if (debug)
	    {
		tls_cert_info_free(tci);
	    }
	    msmtp_endsession(&srv, 0);
	    return exitcode_tls(e);
	}
	if (debug)
	{
	    msmtp_print_tls_cert_info(tci);
	    tls_cert_info_free(tci);
	}
	/* initialize again */
	if ((e = smtp_init(&srv, acc->domain, msg, errstr)) != SMTP_EOK)
	{
	    msmtp_endsession(&srv, 1);
	    return exitcode_smtp(e);
	}
    }
#endif /* HAVE_SSL */

    if (!(srv.cap.flags & SMTP_CAP_ETRN))
    {
	snprintf(errstr, errstr_bufsize,
		"the SMTP server does not support Remote Message Queue Starting");
	msmtp_endsession(&srv, 1);
	return EX_UNAVAILABLE;
    }

    /* authenticate */
    if (acc->auth_mech)
    {
	if (!(srv.cap.flags & SMTP_CAP_AUTH))
	{
	    snprintf(errstr, errstr_bufsize, 
		    "the SMTP server does not support authentication");
	    msmtp_endsession(&srv, 1);
	    return EX_UNAVAILABLE;
	}
	if ((e = smtp_auth(&srv, acc->host, acc->username, acc->password, 
			acc->ntlmdomain, acc->auth_mech, msg, errstr)) != SMTP_EOK)
	{
	    msmtp_endsession(&srv, 1);
	    return exitcode_smtp(e);
	}
    }
    
    /* send the ETRN request */
    if ((e = smtp_etrn(&srv, rmqs_argument, msg, errstr)) != SMTP_EOK)
    {
	msmtp_endsession(&srv, 1);
	return exitcode_smtp(e);
    }

    /* end session */
    msmtp_endsession(&srv, 1);
    return EX_OK;
}


/*
 * msmtp_serverinfo()
 *
 * Prints information about the SMTP server specified in the account 'acc'.
 * 'errstr' must be 'errstr_bufsize' characters long.
 * If an error occured, 'errstr' may contain a descriptive message (or an empty
 * string) and 'msg' may contain the offending message from the SMTP server (or
 * NULL).
 */

int msmtp_serverinfo(account_t *acc, int debug, list_t **msg, char *errstr)
{
    smtp_server_t srv;
    char *server_greeting = NULL;
    int e;
#ifdef HAVE_SSL
    tls_cert_info_t *tci = NULL;
#endif /* HAVE_SSL */
    
    errstr[0] = '\0';
    *msg = NULL;
    
    /* create a new smtp_server_t */
    srv = smtp_new(debug ? stdout : NULL);

    /* prepare tls */
#ifdef HAVE_SSL
    if (acc->tls)
    {
	tci = tls_cert_info_new();
	if ((e = smtp_tls_init(&srv, acc->tls_key_file, acc->tls_cert_file, 
			acc->tls_trust_file, errstr)) != TLS_EOK)
	{
	    e = exitcode_tls(e);
	    goto error_exit;
	}
    }
#endif /* HAVE_SSL */

    /* connect */
    if ((e = smtp_connect(&srv, acc->host, acc->port, errstr)) != NET_EOK)
    {
	e = exitcode_net(e);
       	goto error_exit;
    }

    /* start tls for ssmtp servers */
#ifdef HAVE_SSL
    if (acc->tls && acc->tls_nostarttls)
    {
	if ((e = smtp_tls(&srv, acc->host, acc->tls_nocertcheck, tci, errstr)) 
		!= TLS_EOK)
	{
	    msmtp_endsession(&srv, 0);
	    e = exitcode_tls(e);
	    goto error_exit;
	}
    }
#endif /* HAVE_SSL */

    /* get greeting */
    if ((e = smtp_get_greeting(&srv, msg, &server_greeting, 
		    errstr)) != SMTP_EOK)
    {
	msmtp_endsession(&srv, 0);
	e = exitcode_smtp(e);
       	goto error_exit;
    }
		    
    /* initialize session */
    if ((e = smtp_init(&srv, acc->domain, msg, errstr)) != SMTP_EOK)
    {
	msmtp_endsession(&srv, 1);
	e = exitcode_smtp(e);
       	goto error_exit;
    }
    
    /* start tls for starttls servers */
#ifdef HAVE_SSL
    if (acc->tls && !acc->tls_nostarttls)
    {
	if (!(srv.cap.flags & SMTP_CAP_STARTTLS))
	{
	    snprintf(errstr, errstr_bufsize, 
		    "the SMTP server does not support TLS via the STARTTLS command");
	    msmtp_endsession(&srv, 1);
	    e = EX_UNAVAILABLE;
	    goto error_exit;
	}
	if ((e = smtp_tls_starttls(&srv, msg, errstr)) != SMTP_EOK)
	{
	    msmtp_endsession(&srv, 1);
	    e = exitcode_smtp(e);
	    goto error_exit;
	}
	if ((e = smtp_tls(&srv, acc->host, acc->tls_nocertcheck, tci, errstr)) 
		!= TLS_EOK)
	{
	    msmtp_endsession(&srv, 0);
	    e = exitcode_tls(e);
	    goto error_exit;
	}
	/* initialize again */
	if ((e = smtp_init(&srv, acc->domain, msg, errstr)) != SMTP_EOK)
	{
	    msmtp_endsession(&srv, 1);
	    e = exitcode_smtp(e);
	    goto error_exit;
	}
    }
#endif /* HAVE_SSL */

    /* end session */
    msmtp_endsession(&srv, 1);

    /* print results */
    printf("SMTP server at %s, port %d: %s\n", acc->host, acc->port, server_greeting);
#ifdef HAVE_SSL
    if (acc->tls)
    {
	msmtp_print_tls_cert_info(tci);
    }
#endif /* HAVE_SSL */
#ifdef HAVE_SSL
    if (srv.cap.flags == 0 && !(acc->tls && !acc->tls_nostarttls))
#else /* not HAVE_SSL */
    if (srv.cap.flags == 0)
#endif /* not HAVE_SSL */
    {
	printf("No special capabilities.\n");
    }
    else
    {
	if (srv.cap.flags & SMTP_CAP_SIZE)
	{
	    if (srv.cap.size == 0)
	    {
		printf("SIZE - Maximum message size is unlimited\n");
	    }
	    else
	    {
		printf("SIZE %ld - Maximum message size is %ld bytes",
		     	srv.cap.size, srv.cap.size);
		if (srv.cap.size > 1024 * 1024)
		{
		    printf(" = %.2f MB", (float)srv.cap.size / (1024.0 * 1024.0));
		}
		else if (srv.cap.size > 1024)
		{
		    printf(" = %.2f KB", (float)srv.cap.size / 1024.0);
		}
		printf("\n");
	    }
       	}
	if (srv.cap.flags & SMTP_CAP_PIPELINING)
	{
    	    printf("PIPELINING - Support for command grouping for faster transmission\n");
       	}
	if (srv.cap.flags & SMTP_CAP_ETRN)
	{
    	    printf("ETRN - Support for RMQS (Remote Message Queue Starting)\n");
       	}
	if (srv.cap.flags & SMTP_CAP_DSN)
	{
    	    printf("DSN - Delivery Status Notifications\n");
       	}
#ifdef HAVE_SSL
	if ((acc->tls && !acc->tls_nostarttls) || (srv.cap.flags & SMTP_CAP_STARTTLS))
#else /* not HAVE_SSL */
        if (srv.cap.flags & SMTP_CAP_STARTTLS)
#endif /* not HAVE_SSL */
	{
    	    printf("STARTTLS - Support for TLS encryption via the STARTTLS command\n");
       	}
	if (srv.cap.flags & SMTP_CAP_AUTH)
	{
    	    printf("AUTH - User authentication methods: ");
    	    if (srv.cap.flags & SMTP_CAP_AUTH_PLAIN)
    	    {
    		printf("PLAIN ");
    	    }
    	    if (srv.cap.flags & SMTP_CAP_AUTH_CRAM_MD5)
    	    {
    		printf("CRAM-MD5 ");
    	    }
    	    if (srv.cap.flags & SMTP_CAP_AUTH_DIGEST_MD5)
    	    {
    		printf("DIGEST-MD5 ");
    	    }
    	    if (srv.cap.flags & SMTP_CAP_AUTH_GSSAPI)
    	    {
    		printf("GSSAPI ");
    	    }
    	    if (srv.cap.flags & SMTP_CAP_AUTH_LOGIN)
    	    {
    		printf("LOGIN ");
    	    }
    	    if (srv.cap.flags & SMTP_CAP_AUTH_NTLM)
    	    {
    		printf("NTLM");
    	    }
    	    printf("\n");
       	}
#ifdef HAVE_SSL
	if ((srv.cap.flags & SMTP_CAP_STARTTLS) && !acc->tls)
#else /* not HAVE_SSL */
	if (srv.cap.flags & SMTP_CAP_STARTTLS)
#endif /* not HAVE_SSL */
	{
    	    printf("This server might advertise more or other capabilities when TLS is active.\n");
       	}
    }

    e = EX_OK;
    
error_exit:
#ifdef HAVE_SSL
    if (tci)
    {
	tls_cert_info_free(tci);
    }
#endif /* HAVE_SSL */
    free(server_greeting);
    return e;
}


/*
 * msmtp_read_recipients()
 *
 * Copies the headers of the mail from 'mailf' to a temporary file with stream 
 * '*tmpfile' and name '*tmpfilename', including the blank line that separates
 * the header from the body of the mail.
 * Extracts all recipients from the To, Cc, and Bcc headers and adds them
 * to 'recipients'.
 * 
 * See RFC2822, section 3 for the format of these headers.
 * 
 * Return codes: EX_OK, EX_IOERR
 */

#define STATE_LINESTART_FRESH		0	/* a new line started; the previous line was not a recipient header */
#define STATE_LINESTART_AFTER_RCPTHDR	1	/* a new line started; the previous line was a recipient header */
#define STATE_OTHER_HDR			2	/* a header we don't care about */
#define STATE_TO			3	/* we saw "^T" */
#define STATE_CC			4	/* we saw "^C" */
#define STATE_BCC1			5	/* we saw "^B" */
#define STATE_BCC2			6	/* we saw "^Bc" */
#define STATE_RCPTHDR_ALMOST		7	/* we saw "^To", "^Cc" or "^Bcc" */
#define STATE_RCPTHDR_DEFAULT		8	/* in_rcpt_hdr and in_rcpt state our position */
#define STATE_RCPTHDR_DQUOTE		9	/* duoble quotes */
#define STATE_RCPTHDR_BRACKETS_START	10	/* entering <...> */
#define STATE_RCPTHDR_IN_BRACKETS	11	/* an address inside <> */
#define STATE_RCPTHDR_PARENTH_START	12	/* entering (...) */
#define STATE_RCPTHDR_IN_PARENTH	13	/* a comment inside () */
#define STATE_RCPTHDR_IN_ADDRESS	14	/* a bare address */
#define STATE_RCPTHDR_BACKQUOTE		15	/* we saw a '\\' */
#define STATE_HEADERS_END		16	/* we saw "^$", the blank line between headers and body */

int msmtp_read_recipients(FILE *mailf, list_t *recipients, 
	FILE **tmpfile, char **tmpfilename, char *errstr)
{
    int c;
    int state = STATE_LINESTART_FRESH;
    int oldstate = STATE_LINESTART_FRESH;
    int backquote_savestate = STATE_LINESTART_FRESH;
    int parentheses_depth = 0;
    int parentheses_savestate = STATE_LINESTART_FRESH;
    int folded_rcpthdr_savestate = STATE_LINESTART_FRESH;
    char *current_recipient = NULL;
    size_t current_recipient_len = 0;
    int forget_current_recipient = 0;
    int finish_current_recipient = 0;
    size_t bufsize = 0;
    /* The buffer that is filled with the current recipient grows by 'bufsize_step'
     * if it gets too small. This value must be at least 2. Wasted charcters are at
     * most (bufsize_step - 1). A value of 10 means low wasted space and a low
     * number of realloc()s per recipient. */
    const size_t bufsize_step = 10; 
    

    *tmpfile = NULL;
    *tmpfilename = NULL;
    
    if (tempfile(PACKAGE_NAME, tmpfilename, tmpfile, errstr) != PATHS_EOK)
    {
	goto error_exit;
    }
    
    for (;;)
    {
	c = fgetc(mailf);
	oldstate = state;
	if (c == EOF)
	{
	    state = STATE_HEADERS_END;
	    if (current_recipient)
		finish_current_recipient = 1;
	}
	else
	{
	    switch (state)
	    {
		case STATE_LINESTART_FRESH:
		    parentheses_depth = 0;
		    if (c == 'T')
			state = STATE_TO;
		    else if (c == 'C')
			state = STATE_CC;
		    else if (c == 'B')
			state = STATE_BCC1;
		    else if (c == '\n')
			state = STATE_HEADERS_END;
		    else
			state = STATE_OTHER_HDR;
		    break;

		case STATE_LINESTART_AFTER_RCPTHDR:
		    if (c != ' ' && c != '\t' && current_recipient)
			finish_current_recipient = 1;
		    if (c == ' ' || c == '\t')
			state = folded_rcpthdr_savestate;
		    else if (c == 'T')
			state = STATE_TO;
		    else if (c == 'C')
			state = STATE_CC;
		    else if (c == 'B')
			state = STATE_BCC1;
		    else if (c == '\n')
			state = STATE_HEADERS_END;
		    else
			state = STATE_OTHER_HDR;
		    break;

		case STATE_OTHER_HDR:
		    if (c == '\n')
			state = STATE_LINESTART_FRESH;
		    break;

		case STATE_TO:
	    	    if (c == 'o')
	    		state = STATE_RCPTHDR_ALMOST;
	    	    else if (c == '\n')
	    		state = STATE_LINESTART_FRESH;
	    	    else
	    		state = STATE_OTHER_HDR;
	    	    break;

		case STATE_CC:
		    if (c == 'c')
			state = STATE_RCPTHDR_ALMOST;
		    else if (c == '\n')
			state = STATE_LINESTART_FRESH;
		    else
			state = STATE_OTHER_HDR;
		    break;

		case STATE_BCC1:
		    if (c == 'c')
			state = STATE_BCC2;
		    else if (c == '\n')
			state = STATE_LINESTART_FRESH;
		    else
			state = STATE_OTHER_HDR;
		    break;

		case STATE_BCC2:
		    if (c == 'c')
			state = STATE_RCPTHDR_ALMOST;
		    else if (c == '\n')
			state = STATE_LINESTART_FRESH;
		    else
			state = STATE_OTHER_HDR;
		    break;

		case STATE_RCPTHDR_ALMOST:
		    if (c == ':')
			state = STATE_RCPTHDR_DEFAULT;
		    else if (c == '\n')
			state = STATE_LINESTART_FRESH;
		    else
			state = STATE_OTHER_HDR;
		    break;

		case STATE_RCPTHDR_DEFAULT:
	    	    if (c == '\n')
	    	    {
	    		if (current_recipient)
	    		    finish_current_recipient = 1;
	    		folded_rcpthdr_savestate = state;
	    		state = STATE_LINESTART_AFTER_RCPTHDR;
	    	    }
	    	    else if (c == '\\')
	    	    {
	    		backquote_savestate = state;
	    		state = STATE_RCPTHDR_BACKQUOTE;
	    	    }
	    	    else if (c == '(')
	    	    {
	    		parentheses_savestate = state;
	    		state = STATE_RCPTHDR_PARENTH_START;
	    	    }
	    	    else if (c == '"')
	    	    {
	    		if (current_recipient)
	    		    forget_current_recipient = 1;
	    		state = STATE_RCPTHDR_DQUOTE;
	    	    }
	    	    else if (c == '<')
	    	    {
	    		if (current_recipient)
	    		    forget_current_recipient = 1;
	    		state = STATE_RCPTHDR_BRACKETS_START;
	    	    }
	    	    else if (c == ' ' || c == '\t')
	    		; /* keep state */
	    	    else if (c == ':')
	    	    {
	    		if (current_recipient)
	    		    forget_current_recipient = 1;
	    	    }
	    	    else if (c == ';' || c == ',')
	    	    {
	    		if (current_recipient)
	    		    finish_current_recipient = 1;
	    	    }
	    	    else
	    	    {
	    		if (current_recipient)
	    		    forget_current_recipient = 1;
	    		state = STATE_RCPTHDR_IN_ADDRESS;
	    	    }
	    	    break;

		case STATE_RCPTHDR_DQUOTE:
	    	    if (c == '\n')
	    	    {
	    		folded_rcpthdr_savestate = state;
	    		state = STATE_LINESTART_AFTER_RCPTHDR;
	    	    }
	    	    else if (c == '\\')
	    	    {
	    		backquote_savestate = state;
	    		state = STATE_RCPTHDR_BACKQUOTE;
	    	    }
	    	    else if (c == '"')
	    		state = STATE_RCPTHDR_DEFAULT;
	    	    break;

		case STATE_RCPTHDR_BRACKETS_START:
		    if (c == '\n')
		    {
		       	folded_rcpthdr_savestate = state;
			state = STATE_LINESTART_AFTER_RCPTHDR;
		    }
		    else if (c == '(')
		    {
		       	parentheses_savestate = state;
			state = STATE_RCPTHDR_PARENTH_START;
		    }
		    else if (c == '>')
			state = STATE_RCPTHDR_DEFAULT;
		    else
			state = STATE_RCPTHDR_IN_BRACKETS;
		    break;

		case STATE_RCPTHDR_IN_BRACKETS:
	    	    if (c == '\n')
	    	    {
	    		folded_rcpthdr_savestate = state;
	    		state = STATE_LINESTART_AFTER_RCPTHDR;
	    	    }
	    	    else if (c == '\\')
	    	    {
	    		backquote_savestate = state;
	    		state = STATE_RCPTHDR_BACKQUOTE;
	    	    }
	    	    else if (c == '(')
	    	    {
	    		parentheses_savestate = state;
	    		state = STATE_RCPTHDR_PARENTH_START;
	    	    }
	    	    else if (c == '>')
	    	    {
	    		finish_current_recipient = 1;
	    		state = STATE_RCPTHDR_DEFAULT;
	    	    }
	    	    break;

		case STATE_RCPTHDR_PARENTH_START:
	    	    if (c == '\n')
	    	    {
	    		folded_rcpthdr_savestate = state;
	    		state = STATE_LINESTART_AFTER_RCPTHDR;
	    	    }
	    	    else if (c == ')')
	    		state = parentheses_savestate;
	    	    else
	    	    {
	    		parentheses_depth++;
	    		state = STATE_RCPTHDR_IN_PARENTH;
	    	    }
	    	    break;

		case STATE_RCPTHDR_IN_PARENTH:
	    	    if (c == '\n')
	    	    {
	    		folded_rcpthdr_savestate = state;
	    		state = STATE_LINESTART_AFTER_RCPTHDR;
	    	    }
	    	    else if (c == '\\')
	    	    {
	    		backquote_savestate = state;
	    		state = STATE_RCPTHDR_BACKQUOTE;
	    	    }
	    	    else if (c == '(')
	    		state = STATE_RCPTHDR_PARENTH_START;
	    	    else if (c == ')')
	    	    {
	    		parentheses_depth--;
	    		if (parentheses_depth == 0)
	    		    state = parentheses_savestate;
	    	    }
	    	    break;

		case STATE_RCPTHDR_IN_ADDRESS:
	    	    if (c == '\n')
	    	    {
	    		folded_rcpthdr_savestate = STATE_RCPTHDR_DEFAULT;
	    		state = STATE_LINESTART_AFTER_RCPTHDR;
	    	    }
	    	    else if (c == '\\')
	    	    {
	    		backquote_savestate = state;
	    		state = STATE_RCPTHDR_BACKQUOTE;
	    	    }
	    	    else if (c == '"')
	    	    {
	    		forget_current_recipient = 1;
	    		state = STATE_RCPTHDR_DQUOTE;
	    	    }
	    	    else if (c == '(')
	    	    {
	    		parentheses_savestate = state;
	    		state = STATE_RCPTHDR_PARENTH_START;
	    	    }
	    	    else if (c == '<')
	    	    {
	    		forget_current_recipient = 1;
	    		state = STATE_RCPTHDR_BRACKETS_START;
	    	    }
	    	    else if (c == ' ' || c == '\t')
	    		state = STATE_RCPTHDR_DEFAULT;
	    	    else if (c == ':')
	    	    {
	    		forget_current_recipient = 1;
	    		state = STATE_RCPTHDR_DEFAULT;
	    	    }
	    	    else if (c == ',' || c == ';')
	    	    {
	    		finish_current_recipient = 1;
	    		state = STATE_RCPTHDR_DEFAULT;
	    	    }
	    	    break;

		case STATE_RCPTHDR_BACKQUOTE:
		    if (c == '\n')
		    {
		       	folded_rcpthdr_savestate = STATE_RCPTHDR_DEFAULT;
			state = STATE_LINESTART_AFTER_RCPTHDR;
		    }
		    else
			state = backquote_savestate;
		    break;
	    }
	}

    	if (c != EOF && fputc(c, *tmpfile) == EOF)
	{
	    snprintf(errstr, errstr_bufsize,
		    "output error while writing the mail headers to %s\n",
		    *tmpfilename);
	    goto error_exit;
	}
	
    	if (forget_current_recipient)
	{
	    /* this was just junk */
	    current_recipient_len = 0;
	    current_recipient = NULL;
	    forget_current_recipient = 0;
	}
	if (finish_current_recipient)
	{
	    /* The current recipient just ended. Add it to the list */
	    current_recipient[current_recipient_len] = '\0';
	    list_insert(recipients, current_recipient);
	    recipients = recipients->next;
	    /* Reset for the next recipient */
	    current_recipient_len = 0;
	    current_recipient = NULL;
	    finish_current_recipient = 0;
	}
	if ((state == STATE_RCPTHDR_IN_ADDRESS || state == STATE_RCPTHDR_IN_BRACKETS)
		&& oldstate != STATE_RCPTHDR_PARENTH_START
		&& oldstate != STATE_RCPTHDR_IN_PARENTH
		&& oldstate != STATE_LINESTART_AFTER_RCPTHDR)
	{
    	    /* Add this character to the current recipient */
	    current_recipient_len++;
	    if (bufsize < current_recipient_len + 1)
	    {
		bufsize += bufsize_step;
	    }
	    current_recipient = xrealloc(current_recipient, bufsize * sizeof(char));
	    /* sanitize characters */
	    if (!iscntrl((unsigned char)c) && !isspace((unsigned char)c))
	    {
		current_recipient[current_recipient_len - 1] = (char)c;
	    }
	    else
	    {
		current_recipient[current_recipient_len - 1] = '_';
	    }
	}
	
	if (state == STATE_HEADERS_END)
	{
	    break;
	}
    }
    
    if (ferror(mailf))
    {
	snprintf(errstr, errstr_bufsize, 
		"input error while reading the mail");
	return EX_IOERR;
    }
    
    return EX_OK;

error_exit:
    if (*tmpfile)
    {
	(void)fclose(*tmpfile);
    }
    if (*tmpfilename)
    {
	(void)remove(*tmpfilename);
	free(*tmpfilename);
    }
    free(current_recipient);
    return EX_IOERR;
}

 
/*
 * msmtp_sendmail()
 *
 * Sends a mail. Returns a value from sysexits.h.
 * If 'read_recipients' is true, recipients are extracted from the To, Cc,
 * and Bcc headers *in addition* to the recipients in 'recipients'. 
 * 'errstr' must be 'errstr_bufsize' characters long.
 * If an error occured, 'errstr' may contain a descriptive message (or an empty
 * string) and 'msg' may contain the offending message from the SMTP server (or
 * NULL).
 * In case of success, 'mailsize' contains the number of bytes of the mail 
 * transferred to the SMTP server. In case of failure, its contents are
 * undefined.
 */

int msmtp_sendmail(account_t *acc, list_t *recipients, int read_recipients, 
	FILE *f, int debug, long *mailsize, 
	list_t **msg, char *errstr)
{
    smtp_server_t srv;
    FILE *tmpfile = NULL;
    char *tmpfilename = NULL;
    int e;
#ifdef HAVE_SSL
    tls_cert_info_t *tci = NULL;
#endif /* HAVE_SSL */
    
    errstr[0] = '\0';
    *msg = NULL;
    
    /* Read recipients from the mail as soon as possible. Important for 
     * error reporting/logging. */
    if (read_recipients)
    {
	if ((e = msmtp_read_recipients(f, list_last(recipients), 
			&tmpfile, &tmpfilename, errstr)) != EX_OK)
	{
	    msmtp_endsession(&srv, 1);
	    e = EX_UNAVAILABLE;
	    goto error_exit;
	}
	if (list_is_empty(recipients))
	{
	    snprintf(errstr, errstr_bufsize, "no recipients found");
	    e = EX_DATAERR;
	    goto error_exit;
	}
    }
    /* create a new smtp_server_t */
    srv = smtp_new(debug ? stdout : NULL);

    /* prepare tls */
#ifdef HAVE_SSL
    if (acc->tls)
    {
	if ((e = smtp_tls_init(&srv, acc->tls_key_file, acc->tls_cert_file, 
			acc->tls_trust_file, errstr)) != TLS_EOK)
	{
	    e = exitcode_tls(e);
	    goto error_exit;
	}
    }
#endif /* HAVE_SSL */

    /* connect */
    if ((e = smtp_connect(&srv, acc->host, acc->port, errstr)) != NET_EOK)
    {
	e = exitcode_net(e);
	goto error_exit;
    }

    /* start tls for ssmtp servers */
#ifdef HAVE_SSL
    if (acc->tls && acc->tls_nostarttls)
    {
	if (debug)
	{
	    tci = tls_cert_info_new();
	}
	if ((e = smtp_tls(&srv, acc->host, acc->tls_nocertcheck, tci, errstr)) 
		!= TLS_EOK)
	{
	    if (debug)
	    {
		tls_cert_info_free(tci);
	    }
	    msmtp_endsession(&srv, 0);
	    e = exitcode_tls(e);
	    goto error_exit;
	}
	if (debug)
	{
	    msmtp_print_tls_cert_info(tci);
	    tls_cert_info_free(tci);
	}
    }
#endif /* HAVE_SSL */

    /* get greeting */
    if ((e = smtp_get_greeting(&srv, msg, NULL, errstr)) != SMTP_EOK)
    {
	msmtp_endsession(&srv, 0);
	e = exitcode_smtp(e);
	goto error_exit;
    }
		    
    /* initialize session */
    if ((e = smtp_init(&srv, acc->domain, msg, errstr)) != SMTP_EOK)
    {
	msmtp_endsession(&srv, 1);
	e = exitcode_smtp(e);
	goto error_exit;
    }

    /* start tls for starttls servers */
#ifdef HAVE_SSL
    if (acc->tls && !acc->tls_nostarttls)
    {
	if (!(srv.cap.flags & SMTP_CAP_STARTTLS))
	{
	    snprintf(errstr, errstr_bufsize,
		    "the SMTP server does not support TLS via the STARTTLS command");
	    msmtp_endsession(&srv, 1);
	    e = EX_UNAVAILABLE;
	    goto error_exit;
	}
	if ((e = smtp_tls_starttls(&srv, msg, errstr)) != SMTP_EOK)
	{
	    msmtp_endsession(&srv, 1);
	    e = exitcode_smtp(e);
	    goto error_exit;
	}
	if (debug)
	{
	    tci = tls_cert_info_new();
	}
	if ((e = smtp_tls(&srv, acc->host, acc->tls_nocertcheck, tci, errstr)) 
		!= TLS_EOK)
	{
	    if (debug)
	    {
		tls_cert_info_free(tci);
	    }
	    msmtp_endsession(&srv, 0);
	    e = exitcode_tls(e);
	    goto error_exit;
	}
	if (debug)
	{
	    msmtp_print_tls_cert_info(tci);
	    tls_cert_info_free(tci);
	}
	/* initialize again */
	if ((e = smtp_init(&srv, acc->domain, msg, errstr)) != SMTP_EOK)
	{
	    msmtp_endsession(&srv, 1);
	    e = exitcode_smtp(e);
	    goto error_exit;
	}
    }
#endif /* HAVE_SSL */

    /* authenticate */
    if (acc->auth_mech)
    {
	if (!(srv.cap.flags & SMTP_CAP_AUTH))
	{
	    snprintf(errstr, errstr_bufsize, 
		    "the SMTP server does not support authentication");
	    msmtp_endsession(&srv, 1);
	    e = EX_UNAVAILABLE;
	    goto error_exit;
	}
	if ((e = smtp_auth(&srv, acc->host, acc->username, acc->password, 
			acc->ntlmdomain, acc->auth_mech, msg, errstr)) != SMTP_EOK)
	{
	    msmtp_endsession(&srv, 1);
	    e = exitcode_smtp(e);
	    goto error_exit;
	}
    }
    
    /* prepare for mail sending */
    if (acc->dsn_return || acc->dsn_notify)
    {
	if (!(srv.cap.flags & SMTP_CAP_DSN))
	{
	    snprintf(errstr, errstr_bufsize, 
		    "the SMTP server does not support DSN");
	    msmtp_endsession(&srv, 1);
	    e = EX_UNAVAILABLE;
	    goto error_exit;
	}
    }
    /* send the envelope */
    if ((e = smtp_send_envelope(&srv, acc->from, recipients, 
		    acc->dsn_notify, acc->dsn_return, msg, errstr)) != SMTP_EOK)
    {
	msmtp_endsession(&srv, 1);
	e = exitcode_smtp(e);
	goto error_exit;
    }
    /* send header and body */
    *mailsize = 0;
    if (read_recipients)
    {
	/* first the headers from the temp file */
	rewind(tmpfile);
	if ((e = smtp_send_mail(&srv, tmpfile, acc->keepbcc, mailsize, errstr)) 
	     	!= SMTP_EOK)
	{
	    msmtp_endsession(&srv, 1);
	    goto error_exit;
	}
	/* then the body from the original file */
	if ((e = smtp_send_mail(&srv, f, 1, mailsize, errstr)) != SMTP_EOK)
	{
	    msmtp_endsession(&srv, 1);
	    e = exitcode_smtp(e);
	    goto error_exit;
	}
    }
    else
    {
	if ((e = smtp_send_mail(&srv, f, acc->keepbcc, mailsize, errstr)) 
	     	!= SMTP_EOK)
	{
	    msmtp_endsession(&srv, 1);
	    e = exitcode_smtp(e);
	    goto error_exit;
	}
    }
    /* end the mail */
    if ((e = smtp_end_mail(&srv, msg, errstr)) != SMTP_EOK)
    {
	msmtp_endsession(&srv, 1);
	e = exitcode_smtp(e);
	goto error_exit;
    }

    /* end session */
    msmtp_endsession(&srv, 1);

    e = EX_OK;

error_exit:
    if (tmpfile)
    {
	fclose(tmpfile);
    }
    if (tmpfilename)
    {
	(void)remove(tmpfilename);
	free(tmpfilename);
    }
    return e;
}


/*
 * print_error()
 *
 * Print an error message
 */

void print_error(char *format, ...)
{
    va_list args;
    fprintf(stderr, "%s: ", prgname);
    va_start(args, format);
    vfprintf(stderr, format, args);
    va_end(args);
    fprintf(stderr, "\n");
}


/*
 * msmtp_sanitize_smtpmsg()
 *
 * Sanitizes all lines of the SMTP server message.
 */

void msmtp_sanitize_smtpmsg(list_t *msg)
{
    while (!list_is_empty(msg))
    {
	msg = msg->next;
	msmtp_sanitize_string((char *)msg->data);
    }
}


/*
 * msmtp_log()
 *
 * Append a log entry to 'acc->logfile' with the following information:
 * - date/time 
 * - host=%s
 * - tls=on|off
 * - auth=on|off
 * - from=%s
 * - recipients=%s,%s,... 
 * - mailsize=%s (only if exitcode == EX_OK)
 * - smtpstatus=%s (only if exitcode != EX_OK and a smtp msg is available)
 * - smtpmsg='%s' (only if exitcode != EX_OK and a smtp msg is available)
 * - errormsg='%s' (only if exitcode != EX_OK and an error msg is available)
 * - exitcode=%s 
 * 'exitcode' must be one of the sysexits.h exitcodes.
 */

void msmtp_log(account_t *acc, list_t *recipients, long mailsize, 
	list_t *errmsg, char *errstr, int exitcode)
{
    FILE *f;
    int i;
    time_t t;
    struct tm *tm;
    size_t s;
    list_t *l;
    char *line;
    int n;
    char *p;
    char *failure_reason;
    /* temporary strings: */
    char time_str[64];
    char mailsize_str[80];	/* sufficient for 256 bit longs */
    const char *exitcode_str;
    char smtpstatus_str[4];
    char *smtperrmsg_str = NULL;
    
    
    /*
     * gather information for the log line
     */
    
    line = NULL;
    /* time */
    if ((t = time(NULL)) < 0)
    {
	failure_reason = strerror(errno);
	goto log_failure;
    }
    if (!(tm = localtime(&t)))
    {
	failure_reason = "cannot convert UTC time to local time";
	goto log_failure;
    }
    if (strftime(time_str, 64, "%b %d %H:%M:%S", tm) > 63)
    {
	failure_reason = "BUG: time_str is too short";
	goto log_failure;
    }
    /* mailsize */
    if (exitcode == EX_OK)
    {
	if (snprintf(mailsize_str, 80, "%ld", mailsize) >= 80)
	{
	    failure_reason = "BUG: mailsize_str is too short";
	    goto log_failure;
	}
    }	
    /* exitcode */
    exitcode_str = exitcode_to_string(exitcode);
    /* smtp status and smtp error message */
    if (exitcode != EX_OK && errmsg)
    {
	(void)snprintf(smtpstatus_str, 4, "%d", smtp_msg_status(errmsg));
	msmtp_sanitize_smtpmsg(errmsg);
	l = errmsg;
	s = 0;
	while (!list_is_empty(l))
	{
	    l = l->next;
	    s += strlen(l->data) + 2;
	}
	s += 1;
	smtperrmsg_str = xmalloc(s * sizeof(char));
	smtperrmsg_str[0] = '\'';
	i = 1;
	l = errmsg;
	while (!list_is_empty(l))
	{
	    l = l->next;
	    p = l->data;
	    while (*p != '\0')
	    {
		smtperrmsg_str[i] = (*p == '\'') ? '?' : *p;
		p++;
		i++;
	    }
	    smtperrmsg_str[i++] = '\\';
	    smtperrmsg_str[i++] = 'n';
	}
	i -= 2;
	smtperrmsg_str[i++] = '\'';
	smtperrmsg_str[i++] = '\0';
    }
    
    /*
     * calculate the length of the log line
     */
    
    s = 0;
    /* time: "%b %d %H:%M:%S " */
    s += strlen(time_str) + 1;
    /* "host=%s " */
    s += 5 + strlen(acc->host) + 1;
    /* "tls=on|off " */
    s += 4 + (acc->tls ? 2 : 3) + 1;
    /* "auth=on|off " */
    s += 5 + (acc->auth_mech ? 2 : 3) + 1;
    /* "from=%s " */
    s += 5 + strlen(acc->from) + 1;
    /* "recipients=%s,%s,... " */
    s += 11;
    l = recipients;
    while (!list_is_empty(l))
    {
	l = l->next;
	s += strlen(l->data) + 1;
    }
    /* "mailsize=%s " */
    if (exitcode == EX_OK)
    {
	s += 9 + strlen(mailsize_str) + 1;
    }
    /* "smtpstatus=%s smtpmsg=%s " */
    if (exitcode != EX_OK && errmsg)
    {
	s += 11 + strlen(smtpstatus_str) + 1 + 8 + strlen(smtperrmsg_str) + 1;
    }
    /* "errormsg='%s' */
    if (exitcode != EX_OK && errstr[0] != '\0')
    {
	s += 10 + strlen(errstr) + 2;
    }
    /* "exitcode=%s" */
    s += 9 + strlen(exitcode_str);
    /* '\0' */
    s++;
    
    line = xmalloc(s * sizeof(char));
    
    /*
     * build the log line
     */
    
    /* line should be long enough */
    n = snprintf(line, s, "%s host=%s tls=%s auth=%s from=%s recipients=", 
	    time_str, acc->host, (acc->tls ? "on" : "off"), 
	    (acc->auth_mech ? "on" : "off"), acc->from);
    if ((size_t)n >= s)
    {
	goto line_too_short;
    }
    s -= (size_t)n;
    p = line + n;
    l = recipients;
    while (!list_is_empty(l))
    {
	l = l->next;
	n = snprintf(p, s, "%s,", (char *)(l->data));
	if ((size_t)n >= s)
	{
	    goto line_too_short;
	}
	s -= (size_t)n;
	p += n;
    }
    /* delete the last ',' */
    *(p - 1) = ' ';
    if (exitcode == EX_OK)
    {
	n = snprintf(p, s, "mailsize=%s ", mailsize_str);
	if ((size_t)n >= s)
	{
	    goto line_too_short;
	}
	s -= (size_t)n;
	p += n;
    }
    if (exitcode != EX_OK && errmsg)
    {
	n = snprintf(p, s, "smtpstatus=%s smtpmsg=%s ", smtpstatus_str, smtperrmsg_str);
	if ((size_t)n >= s)
	{
	    goto line_too_short;
	}
	s -= (size_t)n;
	p += n;
    }
    if (exitcode != EX_OK && errstr[0] != '\0')
    {
	msmtp_sanitize_string(errstr);
	n = snprintf(p, s, "errormsg='%s' ", errstr);
	if ((size_t)n >= s)
	{
	    goto line_too_short;
	}
	s -= (size_t)n;
	p += n;
    }
    n = snprintf(p, s, "exitcode=%s", exitcode_str);
    if ((size_t)n >= s)
    {
	goto line_too_short;
    }
    
    /*
     * write log line to file
     */
    
    if (strcmp(acc->logfile, "-") == 0)
    {
	f = stdout;
    }
    else
    {
	if (!(f = fopen(acc->logfile, "a")))
	{
	    failure_reason = strerror(errno);
	    goto log_failure;
	}
    }
    if ((fputs(line, f) == EOF) || (fputc('\n', f) == EOF))
    {
	failure_reason = "output error";
	goto log_failure;
    }
    if (f != stdout && fclose(f) != 0)
    {
	failure_reason = strerror(errno);
	goto log_failure;
    }

    free(line);
    return;
    
    /*
     * error exit targets
     */
    
line_too_short:
    failure_reason = "BUG: log line too short";
    free(line);
    line = NULL;	

log_failure:
    free(smtperrmsg_str);
    print_error("cannot log to %s: %s", acc->logfile, failure_reason);
    if (line && (line[0] != '\0'))
    {
	print_error("log line was: %s", line);
    }
    free(line);
}


/*
 * The main function.
 * It returns values from sysexits.h (like sendmail does).
 */

/* long options without a corresponding short option */
#define LONGONLYOPT_VERSION 		0
#define LONGONLYOPT_HELP    		1
#define LONGONLYOPT_HOST		2
#define LONGONLYOPT_PORT		3
#define LONGONLYOPT_AUTH		4
#define LONGONLYOPT_USER		5
#define LONGONLYOPT_TLS			6
#define LONGONLYOPT_TLS_TRUST_FILE	7
#define LONGONLYOPT_TLS_KEY_FILE	8
#define LONGONLYOPT_TLS_CERT_FILE	9
#define LONGONLYOPT_TLS_CERTCHECK	10
#define LONGONLYOPT_TLS_STARTTLS	11
#define LONGONLYOPT_DOMAIN		12
#define LONGONLYOPT_KEEPBCC		13
#define LONGONLYOPT_RMQS		14

int main(int argc, char *argv[])
{
    /* the configuration */
    int print_version;
    int print_help;
    int print_conf;
    int debug;
    int pretend;
    int read_recipients;
    /* mode of operation */
    int sendmail;
    int serverinfo;
    int rmqs;
    char *rmqs_argument;
    /* account information from the command line */
    account_t *cmdline_account;
    /* account information from the configuration file */
    char *conffile;
    char *account_id;
    list_t *account_list;
    account_t *account;
    /* the recipients of the mail */
    int rcptc;
    char **rcptv;
    list_t *recipients;
    /* error handling */
    int error_code;
    int e;
    char errstr[ERRSTR_BUFSIZE];
    list_t *errmsg;
    list_t *lp;
    /* misc */
#ifdef HAVE_SSL
    int tls_lib_initialized = 0;
#endif
    int net_lib_initialized = 0;
    int i;
    int c;
    long mailsize;
    int option_index;
    struct option options[] =
    {
	{ "version",         no_argument,       0, LONGONLYOPT_VERSION },
	{ "help",            no_argument,       0, LONGONLYOPT_HELP },
	{ "pretend",         no_argument,       0, 'P' },
	{ "debug",           optional_argument, 0, 'd' }, /* ignore opt arg for compat */
	{ "serverinfo",      no_argument,       0, 'S' },
	{ "rmqs",            required_argument, 0, LONGONLYOPT_RMQS },
	{ "file",            required_argument, 0, 'C' },
	{ "account",         required_argument, 0, 'a' },
	{ "host",            required_argument, 0, LONGONLYOPT_HOST },
	{ "port",            required_argument, 0, LONGONLYOPT_PORT },
	{ "from",            required_argument, 0, 'f' },
	{ "auth",            optional_argument, 0, LONGONLYOPT_AUTH },
	{ "user",            required_argument, 0, LONGONLYOPT_USER },
	{ "tls",             optional_argument, 0, LONGONLYOPT_TLS },
	{ "tls-trust-file",  required_argument, 0, LONGONLYOPT_TLS_TRUST_FILE },
	{ "tls-key-file",    required_argument, 0, LONGONLYOPT_TLS_KEY_FILE },
	{ "tls-cert-file",   required_argument, 0, LONGONLYOPT_TLS_CERT_FILE },
	{ "tls-certcheck",   optional_argument, 0, LONGONLYOPT_TLS_CERTCHECK },
	{ "tls-starttls",    optional_argument, 0, LONGONLYOPT_TLS_STARTTLS },
	{ "dsn-notify",      required_argument, 0, 'N' },
	{ "dsn-return",      required_argument, 0, 'R' },
	{ "domain",          required_argument, 0, LONGONLYOPT_DOMAIN },
	{ "keepbcc",         optional_argument, 0, LONGONLYOPT_KEEPBCC },
	{ "logfile",         required_argument, 0, 'X' },
	{ "read-recipients", required_argument, 0, 't' },
	{ 0, 0, 0, 0 }
    };
    
    
    /* process the command line */
    prgname = get_prgname(argv[0]);
    error_code = 0;
    print_version = 0;
    print_help = 0;
    print_conf = 0;
    debug = 0;
    pretend = 0;
    sendmail = 1;
    serverinfo = 0;
    rmqs = 0;
    rmqs_argument = NULL;
    account = NULL;
    cmdline_account = account_new(NULL, NULL);
    conffile = NULL;
    account_list = NULL;
    account_id = NULL;
    recipients = NULL;
    read_recipients = 0;
    for (;;)
    {
	c = getopt_long(argc, argv, "Pd::SC:a:f:N:R:X:tA:B:b:Gh:iL:mnO:o:v", 
		options, &option_index);
	if (c == -1)
	{
	    break;
	}
	switch(c)
	{
	    case LONGONLYOPT_VERSION:
		print_version = 1;
    		sendmail = 0;
		serverinfo = 0;
		break;

	    case LONGONLYOPT_HELP:
		print_help = 1;
		sendmail = 0;
		serverinfo = 0;
		break;

	    case 'P':
		print_conf = 1;
		pretend = 1;
		break;
		
	    case 'd':
		print_conf = 1;
		debug = 1;
		/* ignore an optional argument for sendmail compatibility */
		break;
		
	    case 'S':
		if (rmqs)
		{
		    print_error("cannot use both --serverinfo and --rmqs");
		    error_code = 1;
		}
		else
		{
		    serverinfo = 1;
		    sendmail = 0;
		    rmqs = 0;
		}
		break;

	    case LONGONLYOPT_RMQS:
		if (serverinfo)
		{
		    print_error("cannot use both --serverinfo and --rmqs");
		    error_code = 1;
		}
		else
		{
		    rmqs = 1;
		    rmqs_argument = optarg;
		    sendmail = 0;
		    serverinfo = 0;
		}
		break;
		
	    case 'C':
		free(conffile);
		conffile = xstrdup(optarg);
		break;

	    case 'a':
		if (cmdline_account->host)
		{
		    print_error("cannot use both --host and --account");
		    error_code = 1;
		}
		else
		{
		    account_id = optarg;
		}
		break;

	    case LONGONLYOPT_HOST:
		if (account_id)
		{
		    print_error("cannot use both --host and --account");
		    error_code = 1;
		}
		else
		{
		    free(cmdline_account->host);
		    cmdline_account->host = xstrdup(optarg);
		    cmdline_account->mask |= ACC_HOST;
		}
		break;
		
	    case LONGONLYOPT_PORT:
		cmdline_account->port = atoi(optarg);
		if (cmdline_account->port < 1 || cmdline_account->port > 65535)
		{
		    print_error("invalid port number %d", cmdline_account->port);
		    error_code = 1;
		}
		cmdline_account->mask |= ACC_PORT;
		break;		
		
	    case 'f':
		free(cmdline_account->from);
		cmdline_account->from = xstrdup(optarg);
		cmdline_account->mask |= ACC_FROM;
		break;

	    case LONGONLYOPT_AUTH:
		free(cmdline_account->auth_mech);
		if (optarg)
		{
		    if (is_off(optarg))
		    {
			cmdline_account->auth_mech = NULL;
		    }
		    else if (is_on(optarg))
		    {
			cmdline_account->auth_mech = xstrdup("");
		    }
		    else if (check_auth_arg(optarg) == 0)
		    {
			cmdline_account->auth_mech = xstrdup(optarg);
		    }
		    else
		    {
			print_error("invalid argument %s for %s", optarg, "--auth");
			error_code = 1;
		    }
		}
		else
		{
		    cmdline_account->auth_mech = xstrdup("");
		}
		cmdline_account->mask |= ACC_AUTH_MECH;
		break;

	    case LONGONLYOPT_USER:
		free(cmdline_account->username);
		if (*optarg == '\0')
		{
		    cmdline_account->username = NULL;
		}
		else
		{
		    cmdline_account->username = xstrdup(optarg);
		}
		cmdline_account->mask |= ACC_USERNAME;
		break;

	    case LONGONLYOPT_TLS:
		if (optarg)
		{
		    if (is_on(optarg))
		    {
			cmdline_account->tls = 1;
		    }
		    else if (is_off(optarg))
		    {
			cmdline_account->tls = 0;
		    }
		    else
		    {
			print_error("invalid argument %s for %s", optarg, "--tls");
			error_code = 1;
		    }
		}
		else
		{
		    cmdline_account->tls = 1;
		}
		cmdline_account->mask |= ACC_TLS;
		break;

	    case LONGONLYOPT_TLS_TRUST_FILE:
		free(cmdline_account->tls_trust_file);
		if (*optarg == '\0')
		{
		    cmdline_account->tls_trust_file = NULL;
		}
		else
		{
		    cmdline_account->tls_trust_file = xstrdup(optarg);
		}
		cmdline_account->mask |= ACC_TLS_TRUST_FILE;
		break;

	    case LONGONLYOPT_TLS_KEY_FILE:
		free(cmdline_account->tls_key_file);
		if (*optarg == '\0')
		{
		    cmdline_account->tls_key_file = NULL;
		}
		else
		{
		    cmdline_account->tls_key_file = xstrdup(optarg);
		}
		cmdline_account->mask |= ACC_TLS_KEY_FILE;
		break;

	    case LONGONLYOPT_TLS_CERT_FILE:
		free(cmdline_account->tls_cert_file);
		if (*optarg == '\0')
		{
		    cmdline_account->tls_cert_file = NULL;
		}
		else
		{
		    cmdline_account->tls_cert_file = xstrdup(optarg);
		}
		cmdline_account->mask |= ACC_TLS_CERT_FILE;
		break;

	    case LONGONLYOPT_TLS_CERTCHECK:
		if (optarg)
		{
		    if (is_on(optarg))
		    {
			cmdline_account->tls_nocertcheck = 0;
		    }
		    else if (is_off(optarg))
		    {
			cmdline_account->tls_nocertcheck = 1;
		    }
		    else
		    {
			print_error("invalid argument %s for %s", optarg, "--tls-certcheck");
			error_code = 1;
		    }
		}
		else
		{
		    cmdline_account->tls_nocertcheck = 0;
		}
		cmdline_account->mask |= ACC_TLS_NOCERTCHECK;
		break;

	    case LONGONLYOPT_TLS_STARTTLS:
		if (optarg)
		{
		    if (is_on(optarg))
		    {
			cmdline_account->tls_nostarttls = 0;
		    }
		    else if (is_off(optarg))
		    {
			cmdline_account->tls_nostarttls = 1;
		    }
		    else
		    {
			print_error("invalid argument %s for %s", optarg, "--tls-starttls");
			error_code = 1;
		    }
		}
		else
		{
		    cmdline_account->tls_nostarttls = 0;
		}
		cmdline_account->mask |= ACC_TLS_NOSTARTTLS;
		break;

	    case 'N':
		free(cmdline_account->dsn_notify);
		if (*optarg == '\0' || is_off(optarg))
		{
		    cmdline_account->dsn_notify = NULL;
		}
		else if (check_dsn_notify_arg(optarg) == 0)
		{
		    cmdline_account->dsn_notify = xstrdup(optarg);
		}
		else
		{
		    print_error("invalid argument %s for %s", optarg, "--dsn-notify");
	    	    error_code = 1;
		}
		cmdline_account->mask |= ACC_DSN_NOTIFY;
		break;

	    case 'R':
		/* be compatible to both sendmail and the dsn_notify command */
		free(cmdline_account->dsn_return);
		if (*optarg == '\0' || is_off(optarg))
		{
		    cmdline_account->dsn_return = NULL;
		}
		else if (strcmp(optarg, "hdrs") == 0 || strcmp(optarg, "headers") == 0)
		{
		    cmdline_account->dsn_return = xstrdup("HDRS");
		}
		else if (strcmp(optarg, "full") == 0)
		{
		    cmdline_account->dsn_return = xstrdup("FULL");
		}
		else
		{
		    print_error("invalid argument %s for %s", optarg, "--dsn-return");
	    	    error_code = 1;
		}
		cmdline_account->mask |= ACC_DSN_RETURN;
		break;
		
	    case LONGONLYOPT_DOMAIN:
		free(cmdline_account->domain);
		cmdline_account->domain = xstrdup(optarg);
		cmdline_account->mask |= ACC_DOMAIN;
		break;

	    case LONGONLYOPT_KEEPBCC:
		if (optarg)
		{
		    if (is_on(optarg))
		    {
			cmdline_account->keepbcc = 1;
		    }
		    else if (is_off(optarg))
		    {
			cmdline_account->keepbcc = 0;
		    }
		    else
		    {
			print_error("invalid argument %s for %s", optarg, "--keepbcc");
			error_code = 1;
		    }
		}
		else
		{
		    cmdline_account->keepbcc = 1;
		}
		cmdline_account->mask |= ACC_KEEPBCC;
		break;

	    case 'X':
		free(cmdline_account->logfile);
		if (*optarg == '\0')
		{
		    cmdline_account->logfile = NULL;
		}
		else
		{
		    cmdline_account->logfile = xstrdup(optarg);
		}
		cmdline_account->mask |= ACC_LOGFILE;
		break;

	    case 't':
		read_recipients = 1;
		break;

	    case 'b':
		/* only m makes sense */
		if (strcmp(optarg, "m") != 0)
		{
		    print_error("unsupported operation mode b%s", optarg);
		    error_code = 1;
		}		    
		break;

	    case 'A':
	    case 'B':
	    case 'G':
	    case 'h':
	    case 'i':
	    case 'L':
	    case 'm':
	    case 'n':
	    case 'O':
	    case 'o':
	    case 'v':
		break;
		
	    /* unknown option */
	    default:
		error_code = 1;
		break;
	}
	if (error_code)
	{
	    break;
	}
    }
    if (error_code)
    {
	error_code = EX_USAGE;
	goto exit;
    }

    if (print_version)
    {
	printf("%s version %s\nTLS/SSL library: %s\n"
		"Authentication library: %s\n"
		"Supported authentication methods: ",
		PACKAGE_NAME, VERSION,
#ifdef HAVE_GNUTLS
		"GnuTLS"
#elif defined (HAVE_OPENSSL)
		"OpenSSL"
#else
		"none"
#endif
		,
#ifdef USE_GSASL
		"GNU SASL"
#else
		"built-in"
#endif /* USE_GSASL */
	      );
	if (smtp_client_supports_authmech("PLAIN"))
	{
	    printf("plain ");
	}
	if (smtp_client_supports_authmech("CRAM-MD5"))
	{
	    printf("cram-md5 ");
	}
	if (smtp_client_supports_authmech("DIGEST-MD5"))
	{
	    printf("digest-md5 ");
	}
	if (smtp_client_supports_authmech("GSSAPI"))
	{
	    printf("gssapi ");
	}
	if (smtp_client_supports_authmech("LOGIN"))
	{
	    printf("login ");
	}
	if (smtp_client_supports_authmech("NTLM"))
	{
	    printf("ntlm");
	}
	printf("\n\nCopyright (C) 2004, 2005 Martin Lambers and others.\n"
		"This is free software; see the source for copying conditions.  There is NO\n"
		"warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n");
    }
    if (print_help)
    {
	printf("USAGE:\n\n"
		"Sendmail mode (default):\n"
		"  %s [option...] [--] recipient...\n"
		"  %s [option...] -t [--] [recipient...]\n"
		"  Read a mail from standard input and transmit it to an SMTP server.\n"
		"Server information mode:\n"
		"  %s [option...] --serverinfo\n"
		"  Print information about an SMTP server.\n"
		"Remote Message Queue Starting mode:\n"
		"  %s [option...] --rmqs=host|@domain|#queue\n"
		"  Send a Remote Message Queue Starting request to an SMTP server\n",
		prgname, prgname, prgname, prgname);
	printf("\nOPTIONS:\n\n"
		"General options:\n"
		"  --version                   print version\n"
		"  --help                      print help\n"
		"  -P, --pretend               print configuration info and exit\n"
		"  -d, --debug                 print debugging information\n"
		"Changing the mode of operation:\n"
		"  -S, --serverinfo            print information about the SMTP server\n"
		"  --rmqs=host|@domain|#queue  send a Remote Message Queue Starting request\n");
	printf("Configuration options:\n"
		"  -C, --file=filename         set configuration file\n"
	        "  -a, --account=id            use the specified account from a configuration\n"
		"                              file; command line settings override its settings\n"
		"  --host=hostname             set SMTP server, use only command line settings;\n"
		"                              do not use any configuration file data\n");
	printf("  --port=number               set port number\n"
	        "  --auth[=on|off|mech]        enable/disable authentication and optionally\n"
		"                              choose the method\n"
		"  --user=[username]           set/unset user name for authentication\n"
		"  --tls[=on|off]              enable/disable TLS encryption\n"
		"  --tls-trust-file=[file]     set/unset trust file for TLS\n");
	printf("  --tls-key-file=[file]       set/unset private key file for TLS\n"
		"  --tls-cert-file=[file]      set/unset private cert file for TLS\n"
	        "  --tls-certcheck[=on|off]    enable/disable server certificate checks for TLS\n"
		"  --tls-starttls[=on|off]     enable/disable STARTTLS for TLS\n"
		"  --domain=string             set the argument of SMTP EHLO command\n");
	printf("Options specific to sendmail mode:\n"
		"  -f, --from=address          set envelope from address\n"
		"  -N, --dsn-notify=[off|cond] set/unset DSN notification conditions\n"
		"  -R, --dsn-return=[off|ret]  set/unset DSN notification amount\n"
		"  --keepbcc[=on|off]          enable/disable preservation of the Bcc header\n"
	        "  -X, --logfile=[file]        set/unset log file\n"
	        "  -t, --read-recipients       read additional recipients from the mail\n"
		"  --                          end of options\n");
	printf("Accepted but ignored: -A, -B, -bm, -G, -h, -i, -L, -m, -n, -O, -o, -v\n"
		"\nReport bugs to <%s>.\n", PACKAGE_BUGREPORT);
    }
    
    if (!sendmail && !serverinfo && !rmqs && !print_conf)
    {
	error_code = EX_OK;
	goto exit;
    }

    /* get the list of recipients */
    rcptc = argc - optind;
    rcptv = &(argv[optind]);
    if (rcptc == 0 && sendmail && !read_recipients)
    {
	print_error("no recipients given");
	error_code = EX_USAGE;
	goto exit;
    }
    if (rcptc != 0 && (serverinfo || rmqs))
    {
	print_error("too many arguments");
	error_code = EX_USAGE;
	goto exit;
    }
    recipients = list_new();
    lp = recipients;
    for (i = 0; i < rcptc; i++)
    {
	list_insert(lp, xstrdup(rcptv[i]));
	lp = lp->next;
    }

    /* get the account, either from the conffile or from the command line */
    if (!cmdline_account->host)
    {	
	/* read the configuration file */
	if (!conffile)
	{
	    if ((e = get_default_conffile(CONFFILE, &conffile, errstr)) != CONF_EOK)
	    {
		print_error("cannot get default configuration file name: %s", errstr);
		error_code = EX_CONFIG;
		goto exit;
	    }
	}
	if ((e = get_conf(conffile, &account_list, errstr)) != CONF_EOK)
	{
	    print_error("%s: %s", conffile, errstr);
	    error_code = (e == CONF_EIO) ? EX_IOERR : EX_CONFIG;
	    goto exit;
	}
	/* get the account data */
	if (!account_id)
	{
	    if (cmdline_account->from)
	    {
		/* No account was chosen, but the envelope from address is given.
		 * Choose the right account with this address.
		 */
		account = account_copy(find_account_by_envelope_from(
			    account_list, cmdline_account->from));
	    }
	    if (!account)
	    {
		/* No envelope from address or no matching account. Use default. */
		account_id = "default";
	    }
	}
	if (!account && !(account = account_copy(find_account(account_list, account_id))))
	{
	    print_error("no account %s found in configuration file %s",
	    	    account_id, conffile);
	    error_code = EX_CONFIG;
	    goto exit;
	}
	override_account(account, cmdline_account);
    }
    else
    {
	account = account_copy(cmdline_account);
    }
    /* OK, we're using the settings in 'account'. Complete them and check them. */
    set_default_port(account);
    if (check_account(account, errstr) != CONF_EOK)
    {
	if (account->id && account->conffile)
	{
	    print_error("account %s from %s: %s", account->id, 
		    account->conffile, errstr);
	}
	else
	{
	    print_error("%s", errstr);
	}
	error_code = EX_CONFIG;
	goto exit;
    }
    if (sendmail && !account->from)
    {
	print_error("no envelope from address");
	error_code = EX_USAGE;
	goto exit;
    }
    
    /* print configuration */
    if (print_conf)
    {
	if (account->id && account->conffile)
	{
	    printf("using account %s from %s\n", 
		    account->id, account->conffile);
	}
	printf("host            = %s\n"
		"port            = %d\n"
		"domain          = %s\n",
		account->host,
		account->port,
		account->domain);
	printf("auth            = ");
	if (!account->auth_mech)
	{
	    printf("none\n");
	}
	else if (account->auth_mech[0] == '\0')
	{
	    printf("choose\n");
	}
	else
	{
	    printf("%s\n", account->auth_mech);
	}
	printf("user            = %s\n"
		"password        = %s\n"
		"ntlmdomain      = %s\n"
		"tls             = %s\n"
		"tls_trust_file  = %s\n"
		"tls_key_file    = %s\n"
		"tls_cert_file   = %s\n"
		"tls_starttls    = %s\n"
		"tls_certcheck   = %s\n",
		account->username ? account->username : "(not set)",
		account->password ? "*" : "(not set)",
		account->ntlmdomain ? account->ntlmdomain : "(not set)",
		account->tls ? "on" : "off", 
		account->tls_trust_file ? account->tls_trust_file : "(not set)",
		account->tls_key_file ? account->tls_key_file : "(not set)",
		account->tls_cert_file ? account->tls_cert_file : "(not set)",
		account->tls_nostarttls ? "off" : "on",
		account->tls_nocertcheck ? "off" : "on");
	if (sendmail)
	{
	    printf("from            = %s\n"
		    "dsn_notify      = %s\n"
		    "dsn_return      = %s\n"
		    "keepbcc         = %s\n"
		    "logfile         = %s\n",
		    account->from ? account->from : "(not set)",
		    account->dsn_notify ? account->dsn_notify : "(not set)",
		    account->dsn_return ? account->dsn_return : "(not set)",
		    account->keepbcc ? "on" : "off",
		    account->logfile ? account->logfile : "(not set)");
	    printf("reading recipients from the command line %s\n",
		    read_recipients ? "and the mail" : "");
	}
	if (rmqs)
	{
	    printf("RMQS argument   = %s\n", rmqs_argument);
	}
    }
    if (pretend || (!sendmail && !serverinfo && !rmqs))
    {
	error_code = EX_OK;
	goto exit;
    }
    
    /* initialize libraries */
    if ((sendmail || rmqs) /* serverinfo does not use auth */
	    && account->auth_mech && (strcmp(account->auth_mech, "") != 0) 
	    && !smtp_client_supports_authmech(account->auth_mech))
    {
	print_error("support for authentication method %s is not compiled in", 
		account->auth_mech);
	error_code = EX_UNAVAILABLE;
	goto exit;
    }
    if ((e = net_lib_init(errstr)) != NET_EOK)
    {
	print_error("cannot initialize networking: %s", errstr);
	error_code = EX_SOFTWARE;
	goto exit;	
    }
    net_lib_initialized = 1;
    if (account->tls)
    {
#ifdef HAVE_SSL
	if ((e = tls_lib_init(errstr)) != TLS_EOK)
	{
	    print_error("cannot initialize TLS library: %s", errstr);
	    error_code = EX_SOFTWARE;
	    goto exit;
	}
	tls_lib_initialized = 1;
#else /* not HAVE_SSL */
	print_error("support for TLS is not compiled in");
	error_code = EX_UNAVAILABLE;
	goto exit;
#endif /* not HAVE_SSL */
    }

    /* do the work */
    if (sendmail)
    {
	if ((error_code = msmtp_sendmail(account, recipients,
			read_recipients, stdin, debug, &mailsize, &errmsg, 
			errstr)) != EX_OK)
	{
	    if (errstr[0] != '\0')
	    {
		print_error("%s", errstr);
	    }
	    if (errmsg)
	    {
		msmtp_sanitize_smtpmsg(errmsg);
		lp = errmsg;
		while (!list_is_empty(lp))
		{
		    lp = lp->next;
		    print_error("SMTP server message: %s", lp->data);
		}
	    }		    
	    if (account->id && account->conffile)
	    {
		print_error("could not send mail (account %s from %s)", 
			account->id, account->conffile);
	    }
	    else
	    {
    		print_error("could not send mail");
	    }
	}
	if (account->logfile)
	{
	    msmtp_log(account, recipients, mailsize, errmsg, errstr, error_code);
	}
    }
    else if (serverinfo)
    {
	if ((error_code = msmtp_serverinfo(account, debug, 
			&errmsg, errstr)) != EX_OK)
	{
	    if (errstr[0] != '\0')
	    {
		print_error("%s", errstr);
	    }
	    if (errmsg)
	    {
		msmtp_sanitize_smtpmsg(errmsg);
		lp = errmsg;
		while (!list_is_empty(lp))
		{
		    lp = lp->next;
		    print_error("SMTP server message: %s", lp->data);
		}
	    }		    
	}
    }
    else /* rmqs */
    {
	if ((error_code = msmtp_rmqs(account, debug, rmqs_argument, 
			&errmsg, errstr)) != EX_OK)
	{
	    if (errstr[0] != '\0')
	    {
		print_error("%s", errstr);
	    }
	    if (errmsg)
	    {
		msmtp_sanitize_smtpmsg(errmsg);
		lp = errmsg;
		while (!list_is_empty(lp))
		{
		    lp = lp->next;
		    print_error("SMTP server message: %s", lp->data);
		}
	    }		    
	}
    }

    
exit:

    /* clean up */
#ifdef HAVE_SSL
    if (tls_lib_initialized) 
    {
	tls_lib_deinit();
    }
#endif /* HAVE_SSL */
    if (net_lib_initialized)
    {
	net_lib_deinit();
    }
    free(conffile);
    if (account_list)
    {
	list_xfree(account_list, account_free);
    }
    account_free(account);
    account_free(cmdline_account);
    if (recipients)
    {
	list_xfree(recipients, free);
    }

    return error_code;
}
